from django.apps import apps
from django.db.models.base import ModelBase
from .exceptions import JSONFieldModelTypeError, JSONFieldModelError, JSONFieldValueError


class JSONField:
    types = (int, float, str, bool)

    def __init__(self, field_type=None, required=False, default=None, blank=True, model=None, field=''):
        if field_type not in self.types:
            raise JSONFieldValueError('field_type should be one of: {}'.format(
                str(self.types).replace("<class '", '').replace("'>", '')))
        self.field_type = field_type
        self.required = required or (not blank)
        self.default = default
        self.blank = blank
        self.key = ''
        self.model = self._set_model(model)
        self.field = field

        if model and not field:
            self.field = 'id'

            # if not self.var_type:
            #    raise ValueError('var_type should be specified')

    def check_type(self, value):
        messages = []
        try:
            _value = self.field_type(value)
        except (TypeError, ValueError):
            messages.append('Could not treat {} value `{}` as {}'.format(
                self.key or '', value, self.field_type))
            _value = None
        return _value, messages

    def validate(self, key, value):
        self.key = key
        messages = []
        if self.required and (value is None):
            messages.append('`{}` value is required'.format(key))
        else:
            if not self.blank and value == '':
                messages.append('`{}` value blank is not allowed'.format(key))
            elif (self.default is not None) and (value is None):
                value = self.default
        if messages:
            value = None
        else:
            value, msg = self.check_type(value)
            messages += msg
        return value, messages

    @staticmethod
    def _set_model(model):
        if not model:
            return None
        elif isinstance(model, ModelBase):
            return model
        elif isinstance(model, str):
            try:
                app, _model = model.split('.')
                model = apps.get_model(app, _model)
            except (ValueError, LookupError):
                raise JSONFieldModelError(
                    'model name does not match pattern <application>.<Model> or model does not exist')
            return model
        else:
            raise JSONFieldModelTypeError('wrong model type (NoneType, string or ModelBase allowed)')



import json
import logging
from json import JSONDecodeError

from django.http import HttpResponse
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from .decorators import permissions
from .permission import IsAuthenticated
from .status import HTTP_200_OK, HTTP_500_INTERNAL_SERVER_ERROR, HTTP_400_BAD_REQUEST
from .exceptions import (
    RequestStructureException,
    ResponseError,
    ResponseBadRequestMsgList,
    ResponseOkMessage,
    ResponseFailMessage,
    ResponseStructureException,
    CustomValidationException,
)

from .fields import JSONField
from .settings import rester_settings

logger = logging.getLogger('django_rester')


class BaseAPIView(View):
    auth = rester_settings['AUTH_BACKEND']()
    request_fields, response_fields = {}, {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request_data = None

    @classmethod
    def get_login_field(cls):
        return cls.auth.settings.get('LOGIN_FIELD') or rester_settings.get('LOGIN_FIELD')

    @classmethod
    def as_view(cls, **kwargs):
        view = super(BaseAPIView, cls).as_view()
        view.cls = cls
        return csrf_exempt(view)

    @property
    def _common_request_response_structure(self):
        # check if request_fields are common to any http method or not
        common_structure = False
        for key in self.request_fields.keys():
            if key in self._allowed_methods():
                common_structure = True
                break
        return common_structure

    def _set_response(self, _response):
        if isinstance(_response, (list, tuple)) and len(_response) == 2:
            response = _response[0]
            status = _response[1]
        else:
            response = _response
            status = HTTP_200_OK
        if isinstance(response, HttpResponse):
            result = response
        else:
            try:
                pure_response = json.dumps(response)
                content_type = 'application/json'
            except TypeError:
                pure_response = str(response)
                status = HTTP_500_INTERNAL_SERVER_ERROR
                content_type = 'text/plain'
            result = HttpResponse(pure_response, content_type=content_type, status=status)
            result = self._set_cors(result)
        return result

    def _data_validate(self, method, data, fields, exception, exception_message):
        if fields == {}:
            return data, []
        if self._common_request_response_structure:
            structure = fields.get(method, None)
        else:
            structure = fields
        if not structure and method not in rester_settings.get('FIELDS_CHECK_EXCLUDED_METHODS', []):
            raise exception(exception_message)
        structured_data, messages = self._check_json_field(data, structure)
        if not messages:
            try:
                structured_data = self.custom_validation(structured_data)
                assert structured_data is not None, '.custom_validation() should return validated structured data'
            except (AssertionError, CustomValidationException) as exc:
                messages = ['{}'.format(exc)]
        if fields is self.response_fields and rester_settings.get('SOFT_RESPONSE_VALIDATION', False):
            structured_data = self._add_filtered_data(data, structured_data)
        return structured_data, messages

    def _add_filtered_data(self, data, structured_data):
        # recursive function, validates response_data by response_fields
        value = None
        if isinstance(data, dict):
            for data_key, data_value in data.items():
                structured_value = structured_data.get(data_key, None)
                correct_value = structured_value if structured_value else data_value
                val = self._add_filtered_data(data.get(data_key, None), correct_value)
                if val is not None:
                    if value is None:
                        value = {}
                    value[data_key] = val

        elif isinstance(data, (list, tuple)):
            for item in data:
                val = self._add_filtered_data(item, structured_data[0])
                if val is not None:
                    if value is None:
                        value = []
                    value.append(val)
        else:
            value = structured_data if structured_data else data
        return value

    def _check_json_field(self, data, structure, key='', messages=None):
        # recursive function, validates request_data by request_fields
        value = None
        if messages is None:
            messages = []
        if isinstance(structure, JSONField):
            value, msg = structure.validate(key, data)
            messages += msg
        elif isinstance(structure, dict):
            if not isinstance(data, dict):
                messages.append('{} should be a dict instance'.format(key))
            else:
                for sub_key, sub_structure in structure.items():
                    val, msg = self._check_json_field(data.get(sub_key, None), sub_structure, sub_key)
                    if val is not None:
                        if value is None:
                            value = {}
                        value[sub_key] = val
                    else:
                        messages += msg
        elif isinstance(structure, (list, tuple)):
            if not isinstance(data, (list, tuple)):
                messages.append('{} should be a list or a tuple instance'.format(key))
            else:
                for item in data:
                    val, msg = self._check_json_field(item, structure[0], key)
                    if val is not None:
                        if value is None:
                            value = []
                        value.append(val)
                    else:
                        messages += msg
        return value, messages

    def _set_request_data(self, request):
        request_data, messages = None, []
        method = request.method
        if method in self._allowed_methods():
            try:
                if method == 'GET':
                    request_data = json.loads(json.dumps(request.GET)) if request.GET else {}
                elif method in ('POST', 'PUT', 'PATCH'):
                    request_data = json.loads(request.body.decode('utf-8')) if request.body else {}
                elif method in ('OPTIONS', 'HEAD'):
                    request_data = {}
                if not isinstance(request_data, (dict, list)):
                    raise JSONDecodeError
            except JSONDecodeError:
                messages.append('Request data is not json serializable')
        return request_data, messages

    def custom_validation(self, structured_data):
        # This method needs for custom validation
        # Only CustomValidationException must be raised here
        return structured_data or {}

    def dispatch(self, request, *args, **kwargs):
        self.request_data, messages = self._set_request_data(request)
        _response, status = [None], None
        if not messages:
            user, messages = self.auth.authenticate(request)
            if not messages and user:
                request.user = user
            self.request_data, messages_ = self._data_validate(
                request.method, self.request_data, self.request_fields, RequestStructureException,
                'request data structure is not valid, check for documentation or leave blank')
            if messages_:
                messages_ = [{"request": messages_}]
            messages += messages_
            method_name = request.method.lower()
            if not messages:
                if method_name in self.http_method_names:
                    handler = getattr(self, method_name, self.http_method_not_allowed)
                else:
                    handler = self.http_method_not_allowed

                # TODO refactor this somehow
                # if method_name in ('options',):
                #     _response = handler(request, *args, **kwargs)
                # else:
                _response = list(self.try_response(handler, request, *args, **kwargs))
                messages_ = _response.pop()
                if messages_:
                    messages_ = [{"response": messages_}]
                messages = messages + messages_
        if messages:
            _response = [self.set_response_structure(data=_response[0], message=messages),
                         HTTP_400_BAD_REQUEST]
        _response = self._set_response(_response)
        return _response

    def _set_cors(self, result):
        cors_access = rester_settings.get('CORS_ACCESS')
        result['Access-Control-Allow-Headers'] = 'Access-Control-Allow-Origin, Content-Type, Authorization'
        # TODO: multiple domains in CORS_ACCESS
        """
        Sounds like the recommended way to do it is to have your server read the Origin header 
        from the client, compare that to the list of domains you would like to allow, 
        and if it matches, echo the value of the Origin header back to the client 
        as the Access-Control-Allow-Origin header in the response.
        """
        if isinstance(cors_access, str):
            result['Access-Control-Allow-Origin'] = cors_access
        elif cors_access is True:
            result['Access-Control-Allow-Origin'] = '*'
        else:
            result['Access-Control-Allow-Headers'] = 'Content-Type, Authorization'
        return result

    def options(self, request, *args, **kwargs):
        result = super().options(request, *args, **kwargs)
        result = self._set_cors(result)
        return result

    def try_response(self, handler, request, *args, **kwargs):
        message = []
        messages = []
        success = None
        data = None
        logger.debug('Request: [{} {}] {}'.format(request.method, request.path, self.request_data))
        try:
            success = True
            data = handler(request, *args, **kwargs)
            if isinstance(data, HttpResponse):
                response_status = data.status_code
            else:
                response_status = HTTP_200_OK
                if isinstance(data, tuple) and len(data) == 2:
                    response_status = data[1]
                    data = data[0]
                data, messages = self._data_validate(request.method, data, self.response_fields,
                                                     ResponseStructureException,
                                                     'response data structure is not valid, check for documentation or leave blank')
        except ResponseError as err:
            logger.exception('Error in handler for [{}]'.format(request.path))
            response_status = err.response_status
            if isinstance(err, ResponseBadRequestMsgList):
                message = err.messages
            elif isinstance(err, (ResponseOkMessage, ResponseFailMessage)):
                response_status = err.response_status
                message = err.message
                data = err.data
                success = isinstance(err, ResponseOkMessage)
            else:
                message = '{}'.format(err)
        except Exception as err:
            logger.exception('Error in handler for [{}]'.format(request.path))
            message = '{}'.format(err)
            response_status = HTTP_500_INTERNAL_SERVER_ERROR
        if success is not None:
            success = 200 <= response_status <= 299
        _response = self.set_response_structure(data, success, message)
        logger.debug('Response: [{}] {}'.format(response_status, _response))
        return _response, response_status, messages

    def set_response_structure(self, data=None, success=True, message=None):
        if isinstance(data, HttpResponse):
            _response = data
        else:
            response_structure = rester_settings.get('RESPONSE_STRUCTURE', {})
            if response_structure:
                res_data = {'success': success,
                            'message': message or [],
                            'data': data,
                            }
                str_data = dict(response_structure)
                for item in str_data.keys():
                    if str_data[item] in res_data:
                        str_data[item] = res_data[item]
                _response = str_data
            else:
                _response = data
        return _response


class Login(BaseAPIView):
    response_fields = {"POST": {"token": JSONField(required=True, field_type=str)}}
    request_fields = {"POST": {BaseAPIView.get_login_field(): JSONField(required=True, field_type=str),
                               "password": JSONField(required=True, field_type=str)}}

    def post(self, request):
        data, status = self.auth.login(request, self.request_data)
        return data, status


class Logout(BaseAPIView):
    @permissions(IsAuthenticated)
    def get(self, request):
        data, status = self.auth.logout(request, self.request_data)
        return data, status



import datetime
import jwt
from django.contrib.auth import authenticate
from django.contrib.auth import get_user_model
from django_rediser import RedisStorage
from ..status import HTTP_200_OK, HTTP_401_UNAUTHORIZED
from .settings import rester_jwt_settings
from django_rester.exceptions import ResponseAuthError

redis_db = None
if isinstance(rester_jwt_settings['USE_REDIS'], int) and not isinstance(rester_jwt_settings['USE_REDIS'], bool):
    redis_db = rester_jwt_settings['USE_REDIS']


class BaseAuth:
    _key = '_token'
    _rs = RedisStorage(db=redis_db)
    settings = rester_jwt_settings

    @classmethod
    def _set_payload(cls, user):
        exp = datetime.datetime.timestamp(
            datetime.datetime.now() + datetime.timedelta(seconds=cls.settings['EXPIRE']))
        payload = {item: getattr(user, item, None) for item in cls.settings['PAYLOAD_LIST'] if
                   hasattr(user, item)}
        payload.update({"exp": exp})
        return payload

    def _push_token(self, token):
        self._rs.sadd(self._key, token)

    def _rem_token(self, token):
        return self._rs.srem(self._key, token)

    def _is_member(self, token):
        return self._rs.sismember(self._key, token)

    @classmethod
    def _get_token(cls, request):
        token, messages = None, []
        token = str(request.META.get(cls.settings['AUTH_HEADER'], ''))
        if token:
            if cls.settings['AUTH_HEADER_PREFIX'] and token.startswith(
                    cls.settings['AUTH_HEADER_PREFIX']):
                token = token[len(cls.settings['AUTH_HEADER_PREFIX']):].lstrip()
            else:
                messages.append('Wrong token prefix')
        return token, messages

    @staticmethod
    def _get_user(**kwargs):
        user_model = get_user_model()
        user = user_model.objects.get(**kwargs)
        return user

    def _get_user_data(self, token):
        is_member, data, exp_date, user_data, user, messages = True, None, None, {}, None, []
        if self.settings['USE_REDIS']:
            is_member = self._is_member(token)
        if is_member:
            try:
                data = jwt.decode(token, self.settings['SECRET'], algorithms=[self.settings['ALGORITHM']])
            except jwt.DecodeError:
                messages.append('Wrong authentication token')
            except jwt.ExpiredSignatureError:
                messages.append('Authentication token expired')

            if data:
                exp_date = data.pop('exp', None)
                user_data = {item: data.get(item, None) for item in self.settings['PAYLOAD_LIST']}
            if exp_date and user_data and exp_date > datetime.datetime.now().timestamp():
                user = self._get_user(**user_data)
        else:
            messages.append('Authentication token is not valid or expired')

        return user, messages


class Auth(BaseAuth):
    def login(self, request, request_data):
        token = None
        login = request_data.get(self.settings['LOGIN_FIELD'], None)
        password = request_data.get('password', '')
        if login is not None:
            user = authenticate(username=login, password=password)
        else:
            user = None
        if user:
            payload = self._set_payload(user)
            token, status = jwt.encode(payload, self.settings['SECRET'],
                                       algorithm=self.settings['ALGORITHM']).decode('utf-8'), HTTP_200_OK
            encoded = {'token': token}
        else:
            raise ResponseAuthError('Authentication failed')
        if status == HTTP_200_OK and self.settings['USE_REDIS'] and token:
            self._push_token(token)
        return encoded, status

    def logout(self, request, request_data):
        token, messages = self._get_token(request)
        result = None
        if self.settings['USE_REDIS']:
            result = self._rem_token(token)
        if result == 0:
            messages.append('Token not found')
        elif not result:
            messages.append('Simple logout')
        else:
            messages.append('Token logout')

        return messages, HTTP_200_OK

    def authenticate(self, request_data):
        user, messages = None, None
        token, messages = self._get_token(request_data)
        if token:
            user, messages = self._get_user_data(token)
        if messages:
            user = None
        return user, messages

    def register(self, request_data):
        pass



from .permission import BasePermission
from .status import HTTP_401_UNAUTHORIZED


def permissions(*perms):
    def permissions_decorator(f):
        def wrapper(view, request, *args, **kwargs):
            checked, message = True, ''
            for perm_item in perms:
                if issubclass(perm_item, BasePermission):
                    auth_check = perm_item(request)
                    checked, message = auth_check.check, auth_check.message
                    if not checked:
                        break
            if checked:
                data = f(view, request, *args, **kwargs)
            else:
                data = message, HTTP_401_UNAUTHORIZED
            return data

        return wrapper

    return permissions_decorator



from .status import (
    HTTP_500_INTERNAL_SERVER_ERROR,
    HTTP_400_BAD_REQUEST,
    HTTP_401_UNAUTHORIZED,
    HTTP_200_OK,
)


class ResterException(Exception):
    pass


class ResponseError(ResterException):
    response_status = HTTP_500_INTERNAL_SERVER_ERROR


class ResponseBadRequest(ResponseError):
    response_status = HTTP_400_BAD_REQUEST


class ResponseServerError(ResponseError):
    response_status = HTTP_500_INTERNAL_SERVER_ERROR


class ResponseAuthError(ResponseError):
    response_status = HTTP_401_UNAUTHORIZED


class ResponseOkMessage(ResponseError):
    def __init__(self, message='', data=None, status=HTTP_200_OK):
        self.message = message
        self.data = data
        self.response_status = status


class ResponseFailMessage(ResponseError):
    def __init__(self, message='', data=None, status=HTTP_500_INTERNAL_SERVER_ERROR):
        self.message = message
        self.data = data
        self.response_status = status


class ResponseBadRequestMsgList(ResponseError):
    def __init__(self, messages=None, status=HTTP_400_BAD_REQUEST):
        self.response_status = status
        if messages and isinstance(messages, (list, tuple)):
            self.messages = list(messages)
        elif isinstance(messages, str):
            self.messages = [messages]
        else:
            self.messages = []


class JSONFieldError(ResterException):
    # base JSONField exception
    pass


class JSONFieldModelTypeError(JSONFieldError):
    # JSONField exception, raises when type of model parameter is not valid
    pass


class JSONFieldModelError(JSONFieldError):
    # JSONField exception, raises when value of model parameter is not valid
    pass


class JSONFieldTypeError(JSONFieldError):
    # JSONField exception, simple TypeError inside JSONField class
    pass


class JSONFieldValueError(JSONFieldError):
    # JSONField exception, simple ValueError inside JSONField class
    pass


class BaseAPIViewException(Exception):
    # BaseAPIView exception class
    pass


class RequestStructureException(BaseAPIViewException):
    # raise if request structure is invalid
    pass


class ResponseStructureException(RequestStructureException):
    # raise if response structure is invalid
    pass


class CustomValidationException(BaseAPIViewException):
    # raise if there error in custom validation
    pass



class BasePermission:
    def __init__(self, request):
        self.check, self.message = False, ''
        self.request = request

    @staticmethod
    def _get_message(check, messages):
        return messages.get('SUCCESS' if check else 'FAIL', None)


class IsAuthenticated(BasePermission):
    def __init__(self, request):
        super().__init__(request)
        self.check, self.message = self._is_authenticated()

    def _is_authenticated(self):
        check = self.request.user.is_authenticated and not self.request.user.is_anonymous and self.request.user.is_active
        messages = {'SUCCESS': ['Auth OK'], 'FAIL': ['Required credentials are not provided']}
        message = self._get_message(check, messages)
        return check, message


class IsAdmin(BasePermission):
    def __init__(self, request):
        super().__init__(request)
        self.check, self.message = self._is_admin()

    def _is_admin(self):
        check = self.request.user.is_superuser and self.request.user.is_active
        messages = {'SUCCESS': ['Auth OK'], 'FAIL': ['Required credentials are not provided or user is not superuser']}
        message = self._get_message(check, messages)
        return check, message


class AllowAny(BasePermission):
    def __init__(self, request):
        super().__init__(request)
        self.check, self.message = self._allow_any()

    def _allow_any(self):
        check = True
        messages = {'SUCCESS': ['Auth OK']}
        message = self._get_message(check, messages)
        return check, message
